open ExtArray;;
open ExtString;;
open ExtList;;
let (|>) x f = f x;;
let (>>) f g x = g (f x);;
let flip f x y = f y x;;

type move_kind = 
	| Symb of char * char | Else | Always
	| SymbA of char * char * (int -> unit) | ElseA of (unit -> unit) | AlwaysA of (unit -> unit);;

let frame kind = [| [| kind, 3; |]; [||]; [||] |];;
(* special nodes: 0-in, 1-ok, 2-fail *)
let p_char c = [| [| Symb(c,c), 1; Else, 2 |]; [||]; [||] |];;
let p_range c1 c2 = [| [| Symb(c1,c2), 1; Else, 2 |]; [||]; [||] |];;

let f_char c f = [| [| SymbA(c,c,f), 1; Else, 2 |]; [||]; [||] |];;
let f_range c1 c2 f = [| [| SymbA(c1,c2,f), 1; Else, 2 |]; [||]; [||] |];;

let relocate p delta = 
	Array.map (Array.map (fun (kind, idx) -> (kind, idx + delta) )) p;;

let connect p delta ok_state fail_state = 
	let res = relocate p delta in res.(1) <- [| Always, ok_state |]; res.(2) <- [| Always, fail_state |]; res;; 

let (>>>) p1 p2 = 
	let b_start = 3 + Array.length p1 in
	Array.concat [ frame Always; connect p1 3 b_start 2; connect p2 b_start 1 2];;

let p_opt p = 
	Array.append (frame Always) (connect p 3 1 1);;

let f_opt p f = 
	let res = Array.append (frame Always) (connect p 3 1 1) in 
	res.(5) <- [| AlwaysA f, 1 |];
	res;;

let (|||) p1 p2 = 
	let b_start = 3 + Array.length p1 in
	Array.concat [(frame Always); (connect p1 3 1 b_start); (connect p2 b_start 1 2)];;

let p_star p = 
	Array.append (frame Always) (connect p 3 3 1);;

let f_star initf p = 
	Array.append (frame (AlwaysA initf)) (connect p 3 3 1);;

let p_plus p = p >>> p_star p;;

let f_plus initf p =
	let res = p >>> p_star p in
	res.(0).(0) <- (AlwaysA initf, 3); res;;
 
let (>>=) p f = 
	let res = Array.append (frame Always) (connect p 3 1 2) in
	res.(4) <- [| AlwaysA f, 1 |];
	res;;

external identity : 'a -> 'a = "%identity";;

let rec simplify p =	
	let unconditional = 
		Array.enum p |> Enum.mapi (fun i node ->
			if i=0 then None else
			try Some(i, Array.find (function Always,_->true | _->false) node) with Not_found -> None ) 
		|>	Enum.filter_map identity in
	match Enum.peek unconditional with
	| None -> p
	| Some (from, (uncond_kind, goto)) -> 
			let goto' = if goto < from then goto else goto-1 in
			let adjust_move (kind, idx) = (kind, if idx < from then idx else if idx=from then goto' else idx-1) in
			let p' =  p |> Array.enum |> Enum.mapi (fun i node-> (i,node)) |> Enum.filter_map (fun (i,node) ->
									if i = from then None else Some( Array.map adjust_move node )) 
								|> Array.of_enum in
			simplify p';;		

type move_action = Consume | Keep | ConsumeA of (int -> unit) | KeepA of (unit -> unit);;

let tabulate p =
	p |> Array.map (fun node->
		let tmp = Array.make 257 (Keep, -1) in
		let defm = Array.fold_left (fun def (kind, idx) ->
			match kind with
			| Symb(c1, c2) -> let m = (Consume, idx) in for i = int_of_char c1 to int_of_char c2 do tmp.(i)<-m done; def
			| SymbA(c1, c2, f) -> let m = (ConsumeA f, idx) in for i = int_of_char c1 to int_of_char c2 do tmp.(i)<-m done; def
			| Else 
			| Always -> Keep, idx
			| AlwaysA f -> KeepA f, idx
			| ElseA f -> KeepA f, idx			
			) (Keep, -1) node  in
		tmp |> Array.map (fun (k, idx) -> if idx >=0 then k,idx else defm) );;

let optimize tbl = 
	let rec optimove ch (action, next_state) = 
		if action=Keep && next_state!=1 && next_state!=2 then	optimove ch tbl.(next_state).(ch) 
		else action, next_state   in
	Array.mapi (fun state tab -> if state!=1 && state!=2 then Array.mapi optimove tab else tab) tbl;;		
						
let execute success fail tbl str =
	let len = String.length str in
	let rec run state i =
		if state = 1 then success () else
		if state = 2 then fail () else
		let ch = if i<len then int_of_char str.[i] else 256 in 
		match tbl.(state).(ch) with
		| Consume, next_state -> run next_state (i+1) 
		| Keep, next_state -> run next_state i   
		| ConsumeA f, next_state -> f ch; run next_state (i+1) 
		| KeepA f, next_state -> f (); run next_state i   in
	run 0 0;;
		
let prepare success fail p =
	p |> simplify |> tabulate |> optimize |> execute success fail;;
				
let sign = ref 0.0;;
let ivalue = ref 0;;
let fr = ref 0.0;;
let fv = ref 0.0;;
let nop () = ();;
					
let p_float = 
	f_opt (f_char '-' (fun _->sign:=-1.0)) (fun _->sign:=1.0) 
	>>> f_star (fun _->ivalue:=0) (f_range '0' '9' (fun c->ivalue:=!ivalue*10+c-48))
	>>> p_char ',' 
	>>> f_star (fun _->fv:=0.0; fr:=0.1) 
	      (f_range '0' '9' (fun c-> fv:=!fv +. (float_of_int (c-48)) *. !fr; fr := !fr *. 0.1));;

let sum = ref 0.0;;

let p_floats = 
	f_star (fun _->sum:=0.0) 
	       (p_star (p_char ' ') >>> p_float >>= fun _->sum := !sum +. !sign *. (float_of_int !ivalue +. !fv));;


let parse = 
	p_floats |> prepare 
	  (fun ()->Printf.printf "Ok! Result = %f\n" !sum) 
    (fun ()->print_string "fail!\n");; 
	 
let measured f =
  let t0 = Unix.times() in
  let _ = f () in
  let t1 = Unix.times() in
  t1.Unix.tms_utime -. t0.Unix.tms_utime;;

let ic = open_in "numbers_large.txt" in
let str = input_line ic in
let t = measured (fun ()->parse str) in
close_in ic;
Printf.printf " %f s, %f MB/s" t ((float_of_int (String.length str)) /. t /. 1048576.0);;

